BitArray

Struct BitArray 

pub struct BitArray<const N: usize, Word: Unsigned = usize, const WORDS: usize = { N.div_ceil(Word::UBITS) }> { /* private fields */ }
The BitArray Type

Introduction

A BitArray is a fixed sized vector of bit elements stored compactly in an array of unsigned integer words. The default word type is usize.

The type implements the BitStore trait, which provides a rich API for manipulating the bits in the vector. In addition to the many methods defined by the BitStore trait, the BitArray type provides several ways to construct bit-arrays from various sources.

📝

A BitArray packs its elements into a standard array of some unsigned integer type defined by the generic parameter of type Unsigned. The default Word is a usize which, on modern computer systems, will often be a 64-bit unsigned integer. Operations on and between bit-arrays and other objects in the gf2 crate are implemented using bitwise operations on whole underlying words at a time. These are highly optimised in modern CPUs, allowing for fast computation even on large bit-arrays. It also means we never have to worry about overflows or carries as we would with normal integer arithmetic.

In mathematical terms, a bit-array is a vector over GF(2), the simplest Galois-Field with just two elements, usually denoted 0 & 1, as the booleans true & false, or as the bits set & unset. Arithmetic over GF(2) is mod 2, so addition/subtraction becomes the XOR operation while multiplication/division becomes AND.

It is worth noting that by default, a BitArray prints in vector-order. For example, a bit-vector of size four will print as v0v1v2v3 with the elements in increasing index-order with the “least significant” vector element, v0, coming first on the left. This contrasts to the many bit-set types, which usually print in bit-order. The equivalent object in those libraries with say four elements prints as b3b2b1b0 with the least significant bit b0 printed last on the right.

Of course, for many applications, printing in bit-order makes perfect sense. A size four bit-array initialized with the hex number 0x1 will print as 1000. A bit-ordered version prints the same value as 0001, which will be more natural in some settings.

However, our main aim is numerical work, where vector order is more natural. In particular, bit-order is unnatural for matrices over GF(2). It is too confusing to print a matrix in any order other than the one where the (0,0) element is at the top left, and proceed from there.

Methods Overview

The BitArray type implements the BitStore trait and also provides several methods for constructing bit-arrays:

CategoryDescription
Trait RequirementsMethods needed to implement the BitStore trait.
ConstructorsMethods to create bit-arrays with specific properties and fills.

The type also inherits dozens of associated methods from the BitStore trait. These methods fall into categories:

Inherited CategoryDescription
Bit AccessMethods to access individual bit elements in a bit-array.
QueriesMethods to query the overall state of a bit-array.
MutatorsMethods to mutate the overall state of a bit-array.
Copies & FillsMethods to fill a bit-array from various sources.
SlicesMethods to create non-owning views over a part of a bit-array — bit-slices.
Sub-vectorsMethods to clone a piece of a bit-array as a new bit-vector.
RifflingMethods to create vectors that copy a bit-array with interleaved zeros.
Set/Unset IndicesMethods to find the indices of set & unset bits in a bit-array.
IteratorsMethods to create various iterators over a bit-array.
StringificationMethods to create string representations of a bit-array.
Bit ShiftsMethods to shift the bits in a bit-array left or right.
Bitwise OperationsMethods to combine any bit-store and a bit-array using logical operations.
Arithmetic OperatorsMethods to add or subtract any bit-store and a bit-array.
Other FunctionsDot products, convolutions, etc. for bit-stores with bit-arrays.

Trait Requirements

To implement the BitStore trait, the type defines the following seven methods:

MethodDescription
BitArray::lenReturns the number of bits in the bit-array — the generic parameter N.
BitArray::storeProvides read-only access to the first word holding bits in the bit-array.
BitArray::store_mutProvides read-write access to the first word holding bits in the bit-array.
BitArray::offsetReturns the offset in bits from start of word(0) to the bit-array’s first element.
BitArray::wordsThis is always Word::words_needed(self.len()) but cached for efficiency.
BitArray::wordReturns a “word” from the bit-array.
BitArray::set_wordSets the value of a “word” in the bit-array to a passed value.

These methods are trivial to implement for bit-arrays.

The one place where care is needed is in the BitArray::set_word method, which must ensure that any bits beyond the size of the bit-array remain set to zero.

Constructors

The BitArray type provides several constructors to create bit-arrays with specific properties and fills:

Method NameDescription
BitArray::newReturns a bit-array that has N zero elements.
BitArray::zerosReturns a bit-array where all the elements are 0.
BitArray::onesReturns a bit-array where all the elements are 1.
BitArray::constantReturns a bit-array where all the elements are whatever is passed as a value.
BitArray::unitReturns a bit-array where all the elements are zero except for a single 1.
BitArray::alternatingReturns a bit-array where all the elements follow the pattern 101010...
BitArray::from_wordReturns a bit-array filled with bits copied from a Word value.
BitArray::from_fnReturns a bit-array filled with bits set by calling a function for each index.
BitArray::randomReturns a bit-array filled by flipping a fair coin seeded from entropy.
BitArray::random_seededReturns a bit-array with a reproducible fair random fill.
BitArray::random_biased Returns a random bit-array where you set the probability of bits being 1.
BitArray::random_biased_seededReturns a random bit-array where you set the probability of bits being 1 and the RNG’s seed.

Note: We have implemented the Default trait for BitArray to return a zero-sized bit-vector.

Bit Access (Inherited)

The following methods provide access to individual bit elements in the bit-array.

MethodDescription
BitStore::getReturns the value of a single bit element as a read-only boolean.
BitStore::firstReturns the value of the first element in the bit-array.
BitStore::lastReturns the value of the last element in the bit-array.
BitStore::setSets a bit to the given boolean value.
BitStore::flipFlips the value of the bit element at a given index.
BitStore::swapSwaps the values of bit elements at locations i and j.

Queries (Inherited)

The following methods let you query the overall state of a bit-array.

MethodDescription
BitStore::is_emptyReturns true if the bit-array is empty
BitStore::anyReturns true if any bit in the bit-array is set.
BitStore::allReturns true if every bit in the bit-array is set.
BitStore::noneReturns true if no bit in the bit-array is set.
BitStore::count_onesReturns the number of set bits in the bit-array.
BitStore::count_zerosReturns the number of unset bits in the bit-array.
BitStore::leading_zerosReturns the number of leading unset bits in the bit-array.
BitStore::trailing_zerosReturns the number of trailing unset bits in the bit-array.

Mutators (Inherited)

The following methods let you mutate the entire bit-array in a single call.

MethodDescription
BitStore::set_allSets all the bits in the bit-array to the passed value.
BitStore::flip_allFlips the values of all the bits in the bit-array.
BitStore::flippedReturns a new bit-array that is a copy of the bit-vector with all the bits flipped.

Copies and Fills (Inherited)

The following methods let you populate the entire bit-array from multiple sources in a single call.

MethodDescription
BitStore::copy_unsignedCopies bit values from any unsigned value to this bit-array.
BitStore::copy_storeCopies bit values from any source bit-store to this bit-array.
BitStore::copy_fnCopies bit values from a function that returns a boolean for an index.
BitStore::fill_random_biased_seededVery general method to fill the bit-array with random 0’s and 1’s.
BitStore::fill_random_biasedFill the bit-array with random 0’s and 1’s, where the RNG itself is randomly seeded.
BitStore::fill_randomFill the bit-array with random 0’s and 1’s from flips of a fair coin.

Slices (Inherited)

The following methods let you create a BitSlice, which is a non-owning view of some contiguous subset of bits in the bit-array.

MethodDescription
BitStore::sliceReturns a BitSlice encompassing the bits in a half-open range.
BitStore::slice_mutReturns a mutable BitSlice encompassing the bits in a half-open range.

Sub-vectors (Inherited)

The following methods create or fill independent bit-vectors with copies of some contiguous subset of the bits in the bit-array.

MethodDescription
BitStore::subReturns a new BitVector encompassing the bits in a half-open range.
BitStore::split_at_intoFills two bit-vectors with the bits in the ranges [0, at) and [at, len()).
BitStore::split_atReturns two new two bit-vectors with the bits in the ranges [0, at) and [at, len()).

Riffling (Inherited)

We have methods that can interleave (riffle) the bits in a bit-array with zeros.

MethodDescription
BitStore::riffled_intoFills a pre-existing bit-vector with the result of riffling this bit-array.
BitStore::riffledReturns a new bit-vector that is this bit-array with its bits interleaved with zeros.

Set and Unset Bit Indices (Inherited)

The following methods find the indices of set or unset bits in the bit-array.

MethodDescription
BitStore::first_setReturns the index of the first set bit in the bit-array.
BitStore::last_setReturns the index of the last set bit in the bit-array.
BitStore::next_setReturns the index of the next set bit in the bit-array after the passed index.
BitStore::previous_setReturns the index of the previous set bit in the bit-array before the passed index.
BitStore::first_unsetReturns the index of the first unset bit in the bit-array.
BitStore::last_unsetReturns the index of the last unset bit in the bit-array.
BitStore::next_unsetReturns the index of the next unset bit in the bit-array after the passed index.
BitStore::previous_unsetReturns the index of the previous unset bit in the bit-array before the passed index.

Iterators (Inherited)

The following methods create iterators for traversing the bits or underlying words in the bit-array:

MethodDescription
BitStore::bitsReturns a Bits iterator over the bits in the bit-array.
BitStore::set_bitsReturns a SetBits iterator to view the indices of all the set bits.
BitStore::unset_bitsReturns a UnsetBits iterator to view the indices of all the unset bits.
BitStore::store_wordsReturns a Words iterator to view the “words” underlying the bit-array.
BitStore::to_wordsReturns a copy of the “words” underlying the bit-array.

Stringification (Inherited)

The following functions returns a string representation of a bit-array. The string can be in the obvious binary format or a more compact hex format.

MethodDescription
BitStore::to_custom_binary_stringReturns a binary string representation for a bit-array with various customisation parameters.
BitStore::to_binary_stringReturns the simplest binary string representation for a bit-array.
BitStore::to_pretty_stringReturns a “pretty” binary string representation for a bit-array.
BitStore::to_hex_stringReturns a compact hex string representation for a bit-array.
std::string::ToString::to_stringDelegates to BitStore::to_binary_string.
BitStore::describeReturns a multi-line string describing the bit-array in some detail.

Bit Shifts (Inherited)

We have methods to shift the bits in a bit-vector left or right.

MethodsDescription
BitStore::left_shiftLeft shifts in-place.
BitStore::right_shiftRight shifts in-place.
BitStore::left_shiftedCopies the bit-array to a new bit-vector and left shifts that vector.
BitStore::right_shiftedCopies the bit-array to a new bit-vector and right shifts that vector.

Note: We have also implemented the std::ops::ShlAssign, std::ops::ShrAssign, std::ops::Shl, and std::ops::Shr foreign traits to provide operator overloads for the shift operations. Those implementations forward to the associated methods above.

Bitwise Operations (Inherited)

We have methods that combine a bit-array with any other bit-store using the logical operations XOR, AND, and OR.

MethodDescription
BitStore::xor_eqIn-place XOR operation of equal-sized bit-stores: lhs = lhs ^ rhs.
BitStore::and_eqIn-place AND operation of equal-sized bit-stores: lhs = lhs & rhs.
BitStore::or_eqIn-place OR operation of equal-sized bit-stores: lhs = lhs | rhs.
BitStore::xorReturns the XOR of this store with another equal-sized store as a new bit-vector.
BitStore::andReturns the AND of this store with another equal-sized store as a new bit-vector.
BitStore::orReturns the OR of this store with another equal-sized store as a new bit-vector.

Note: We have also implemented the std::ops::BitXorAssign, std::ops::BitAndAssign, std::ops::BitOrAssign, std::ops::BitXor, std::ops::BitAnd, and std::ops::BitOr foreign traits to provide operator overloads for the bit-wise operations. Those implementations forward to the associated methods above.

Arithmetic Operations (Inherited)

In GF(2), the arithmetic operators + and - are both the XOR operator.

MethodDescription
BitStore::plus_eqAdds the passed (equal-sized) rhs bit-store to this bit-vector.
BitStore::minus_eqSubtracts the passed (equal-sized) rhs bit-store from this bit-vector.
BitStore::plusAdds two equal-sized bit-stores and returns the result as a bit-vector.
BitStore::minusSubtracts two equal-sized bit-stores and returns the result as a bit-vector.

Note: We have also implemented the std::ops::AddAssign, std::ops::SubAssign, std::ops::Add, and std::ops::Sub foreign traits to provide operator overloads for the arithmetic operations. Those implementations forward to the associated methods above.

Other Inherited Functions

MethodDescription
BitStore::dotReturns the dot product of two equal-sized bit-stores as a boolean.
BitStore::convolved_withReturns the convolution of two bit-stores as a new bit-vector.

Foreign Traits for Individual Bit-Arrays

We have implemented several foreign traits from the standard library for bit-vectors.

The std::ops::Not trait is implemented for bit-arrays by value and by reference.

Foreign Traits for Pairwise Bit-Array Operations with Other Bit-Stores

We have implemented several foreign traits from the standard library for bit-vectors interacting with other bit-stores.

Implementations

Source

impl<const N: usize, Word: Unsigned, const WORDS: usize> BitArray<N, Word, WORDS>

Constructors for bit-arrays.

pub fn new() -> Self

The default constructor constructs a bit-array with N elements that are all zeros.

Examples
use gf2::*;
let v: BitArray<10, u8> = BitArray::new();
assert_eq!(v.to_string(), "0000000000");
pub fn from_word(word: Word) -> Self

Constructs a bit-array with N elements by repeatedly copying the bits from a single Word instance.

The final copy of word may be truncated and padded with zeros (unused bits are always set to zero in this crate).

Examples
use gf2::*;
let v: BitArray<10, u8> = BitArray::from_word(0b01010101);
assert_eq!(v.to_string(), "1010101010");
pub fn zeros() -> Self

Constructs a bit-array with ``Nelements, all initialized to0`.

Examples
use gf2::*;
let v: BitArray<10, u8> = BitArray::zeros();
assert_eq!(v.to_string(), "0000000000");
pub fn ones() -> Self

Constructs a bit-array with N elements, all initialized to 1.

Examples
use gf2::*;
let v: BitArray<10, u8> = BitArray::ones();
assert_eq!(v.to_string(), "1111111111");
pub fn constant(val: bool) -> Self

Returns a bit-array with N elements, all initialized to the boolean value val.

Examples
use gf2::*;
let v: BitArray<10> = BitArray::constant(true);
assert_eq!(v.to_string(), "1111111111");
let v: BitArray<10> = BitArray::constant(false);
assert_eq!(v.to_string(), "0000000000");
pub fn unit(i: usize) -> Self

Constructs the unit bit-array where only element i of N elements is set to 1.

Panics

Panics if i is greater than or equal to N.

Examples
use gf2::*;
let v: BitArray<10> = BitArray::unit(5);
assert_eq!(v.to_string(), "0000010000");
pub fn alternating() -> Self

Constructs a bit-array with N elements, where the bits alternate between 1 and 0.

The pattern starts with a 1 so e.g. 1010101010101.

Examples
use gf2::*;
let v: BitArray<10> = BitArray::alternating();
assert_eq!(v.to_string(), "1010101010");
pub fn from_fn(f: impl Fn(usize) -> bool) -> Self

Constructs a bit-array with N elements by calling a function f for each bit index.

Examples
use gf2::*;
let v: BitArray<10> = BitArray::from_fn(|i| i % 2 == 0);
assert_eq!(v.to_string(), "1010101010");
Source

impl<const N: usize, Word: Unsigned, const WORDS: usize> BitArray<N, Word, WORDS>

Constructors that set the elements of a bit-array randomly.

pub fn random() -> Self

Constructs a random bit-array with N elements where each bit is set/unset with probability 50/50.

The random number generator is seeded on first use with a scrambled version of the current time so you get different outputs for each run.

See the random_seeded method for a way to get reproducible randomly filled bit-arrays.

Examples
use gf2::*;
let v: BitArray<10> = BitArray::random();
assert_eq!(v.len(), 10);
pub fn random_seeded(seed: u64) -> Self

Constructs a random bit-array with N elements where each bit is set/unset with probability 50/50.

For reproducibility, the random number generator used here is seeded with the specified seed.

Note
  • The generator is reset to the previous seed after the bit-array is constructed.
  • A seed of 0 is taken to seed the generator from a scrambled version of the current time.
Examples
use gf2::*;
let v1: BitArray<1000, u8> = BitArray::random_seeded(42);
let v2: BitArray<1000, u8> = BitArray::random_seeded(42);
assert_eq!(v1, v2);
pub fn random_biased(p: f64) -> Self

Constructs a random bit-array with N elements where each bit is set with probability p.

The random number generator is seeded on first use with a scrambled version of the current time.

Note

Probability p should be in the range [0, 1]. If p is outside this range, the function will return a bit-array with all elements set or unset as appropriate.

Examples
use gf2::*;
let v: BitArray<10> = BitArray::random_biased(0.578);
assert_eq!(v.len(), 10);
pub fn random_biased_seeded(p: f64, seed: u64) -> Self

Constructs a random bit-array with N elements where each bit is set with probability p and the RNG is

Note

Probability p should be in the range [0, 1]. If p is outside this range, the function will return a bit-array with all elements set or unset as appropriate.

Examples
use gf2::*;
let v: BitArray<10> = BitArray::random_biased_seeded(1.2, 42); // All bits set since p > 1
assert_eq!(v.count_ones(), 10);
let u: BitArray<100> = BitArray::random_biased_seeded(0.85, 42); // Using same seed for u and v.
let v: BitArray<100> = BitArray::random_biased_seeded(0.85, 42);
assert_eq!(u, v);
Source

impl<const N: usize, Word: Unsigned, const WORDS: usize> BitArray<N, Word, WORDS>

Vector-matrix multiplication for a BitArray

pub fn dot_matrix(&self, rhs: &BitMatrix<Word>) -> BitVector<Word>

BitArray matrix multiplication as a new [BitVector`].

Panics

Panics if the operands have incompatible dimensions.

Examples
use gf2::*;
let v: BitVector = BitVector::ones(3);
let m: BitMatrix = BitMatrix::identity(3);
assert_eq!((&v * &m).to_string(), "111");